/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Prepaid calling card application
 * 
 * Copyright (C) 2003, Green Tone N.V. and Igor de Rooij
 *
 * $Id: app_prepaid_cid.c,v 1.2 2004/06/19 21:41:44 peternixon Exp $
 *
 * TODO:
 * - better error checks
 * - TEST TEST TEST!
 *
 * Changed:
 * - added connection checking and reconnecting stuff
 * - altered the underlying database layout
 * - changed to cid authentification instead of PIN Code
 * - added PQclear so that the memory gets free
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/app.h>
#include <asterisk/say.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>

#include <asterisk/config.h>
#include <postgresql/libpq-fe.h>

#include <asterisk/translate.h>
#include <asterisk/parking.h>
#include <asterisk/musiconhold.h>
#include <asterisk/callerid.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <asterisk/options.h>

#define DATE_FORMAT "%Y-%m-%d %T"

static char *tdesc = "Prepaid CID Application";

static char *app = "PrepaidCID";

static char *config = "prepaidcid.conf";

static char *synopsis = "Prepaid Authentication, Authorization and Accounting";

static char *descrip =
  "  PrepaidCID: Authenticates a CID and authorizes the called\n"
  " destination and handles accounting\n";

static char *pghostname = NULL, *pgdbname = NULL, *pgdbuser = NULL, *pgpassword = NULL, *pgdbsock = NULL, *pgdbport = NULL;
static char *psqlauthentication = NULL, *psqlcheckforsession = NULL, *psqlauthorization = NULL, *psqlacctstart = NULL, *psqlacctstop = NULL, *psqlsimulcount = NULL;
static int hostname_alloc = 0, dbname_alloc = 0, dbuser_alloc = 0, password_alloc = 0, dbsock_alloc = 0, dbport_alloc = 0;
static int sqlauthentication_alloc = 0, sqlcheckforsession_alloc = 0, sqlauthorization_alloc = 0, sqlacctstart_alloc = 0, sqlacctstop_alloc = 0, sqlsimulcount_alloc = 0;
static int connected = 0;

PGconn *conn;

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

struct calldata {
	char cid[256];
	char destination[256];
	int  credit;
	int tariffid;
	char *active;
	char *rate;
	char *cardid;
	char *tech;
	char *prefix;
	char *hostname;
	int providerid;
	int countryprefixid;
	int tariffrateid;
	int timeout;
};

// Checks if we are connected to the db - if not then it tries to reconnect
static int check_connected(void)
{
	int i;
	PGresult *result;
	
	if (connected) {
		// seems that we are connected - but let us check that
		for (i=0;i<3;i++) {
			// PQstatus is not working as expected - so we are trying a query before PQstatus - hope that helps
			result = PQexec(conn, "SELECT 1");
			PQclear(result);
			if (PQstatus(conn)==CONNECTION_BAD) {
				// seems that we are having troubles with the connection - trying to reset the connection
				// and set connected to false
				connected = 0;
				ast_log(LOG_ERROR, "app_prepaid: connection to the database server %s failed - trying to reconnect (3 times)\n", pghostname);
				PQreset(conn);
				// now retry it (3 times)
				continue;
			} else {
				// Everything is fine - so go on
				connected = 1;
				break;
			}
		}
	} else {
		// We are not connected - so make the initial connection
		conn = PQsetdbLogin(pghostname, pgdbport, NULL, NULL, pgdbname, pgdbuser, pgpassword);
		if (PQstatus(conn) != CONNECTION_BAD) {
			ast_log(LOG_DEBUG,"Successfully connected to Prepaid database.\n");
			connected = 1;
		} else {
			ast_log(LOG_ERROR, "app_prepaid: cannot connect to database server %s.  Calls will not be logged\n", pghostname);
			connected = 0;
		}
	}
	return connected;
}

static int prepaid_ivr_authenticate(struct ast_channel *chan, struct calldata *cd)
{
	int res=0;
	int i=0;
	int retries;
	// char cardnumber[256] = "";
	char *prompt = NULL;
	char sqlcmd[2048] = "";
	PGresult *result;

	// Wait a second before doing anything
	ast_waitfor(chan,5000);
	if (check_connected()) {
		for (retries = 0; retries < 3; retries++) {
			// Check the Callerid
			if (strlen(chan->callerid)==0) {
				// Seems that we don't got a caller id - so auth will fail
				prompt="prepaid-no-cid";
				res = -2;
				break;
			}
			strcpy (cd->cid,chan->callerid);
			sprintf(sqlcmd,psqlauthentication,chan->callerid);
			ast_log(LOG_DEBUG,"app_prepaid: SQL Authenticate command as follows:  %s\n",sqlcmd);
			result = PQexec(conn, sqlcmd);
			if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
				prompt="prepaid-no-aaa"; // tech error
				res = -2;
				break;
			}
			if(PQntuples(result)>0) {
				// There can be more than one result - so check every result
				cd->credit = 0;
				// Basically we think that that it is inactive
				cd->active = "f";
				for (i = 0; i < PQntuples(result); i++) {
					if ((strcmp(PQgetvalue(result,i,2), "t") == 0) && (strcmp(PQgetvalue(result,i,3), "f") == 0)) {
						// Card is active and not expired - so use it
						cd->credit = cd->credit + atoi(PQgetvalue(result,i,0));
						// Set active to true
						cd->active = "t";
						// We take the tariffid from the last card
						cd->tariffid = atoi(PQgetvalue(result,i,1));
						// We take the first active cardid
						if (!cd->cardid) {
							cd->cardid = (char *)malloc(strlen(PQgetvalue(result,i,4))+1);
							strcpy (cd->cardid,PQgetvalue(result,i,4));
						}
					}
				}
				ast_log(LOG_DEBUG,"app_prepaid: SQL Authenticate results:  cid='%s', credit='%d', tariffid='%d', activated='%s'\n",cd->cid,cd->credit,cd->tariffid,cd->active);
			} else {
				prompt="prepaid-auth-failed"; // no results - cid is not activated
				res = -2;
				break;
			}	
			PQclear(result);
			// check credit > 0 / you have zero balance
			if( cd->credit <= 0 ) {
				prompt = "prepaid-zero-balance";
				res = -2;
				break;
			}
	
			// check activated=t / card not active, contact customer support
			if(strcmp(cd->active, "t") != 0 ) {
				// not expired but inactive.. probably not yet sold.. find better prompt
				prompt = "prepaid-card-inactive";
				res = -2;
				break;
			}
	
			// check double logins..
			sprintf(sqlcmd,psqlcheckforsession, cd->cid);
			result = PQexec(conn, sqlcmd);
			if ( PQresultStatus(result) == PGRES_TUPLES_OK) {
				if (PQntuples(result) > 0) {
					// If we have a result - then there is an open session
					prompt="prepaid-card-in-use"; // tech error sql failed
					res = -2;
					break;
				}
			}
			PQclear(result);
			
			break;
		}
	
		if ((retries < 3) && !res) {
			int euros, cents;
			div_t div_result;
			ast_cdr_setaccount(chan, cd->cid);
	
			// say 'you have x euros and x cents'
			div_result = div (cd->credit,100);
	
			euros = div_result.quot;
			cents = div_result.rem;
	
			res = ast_streamfile(chan, "prepaid-you-have", chan->language);
			if (!res)
				res = ast_waitstream(chan, "");
	
			if (euros > 0) {
				res = ast_say_number(chan, euros, "", chan->language, "f");
				if (!res)
					res = ast_waitstream(chan, "");
			}
	
			res = ast_streamfile(chan, "prepaid-euro", chan->language);
			if (!res)
				res = ast_waitstream(chan, "");
	
			if (euros > 0 && cents > 0) {
				res = ast_streamfile(chan, "prepaid-and", chan->language);
				if (!res)
					res = ast_waitstream(chan, "");
			}
	
			if ( cents > 0) {
				res = ast_say_number(chan, cents, "", chan->language, "f");
				if (!res)
					res = ast_waitstream(chan, "");
				res = ast_streamfile(chan, "prepaid-cents", chan->language);
				if (!res)
					res = ast_waitstream(chan, "");
				res = 0;
			}
	
		} else if (res == -2 ) {
			res = ast_streamfile(chan, prompt, chan->language);
			if (!res)
				res = ast_waitstream(chan, "");
			res = -1;
		} else {
			res = ast_streamfile(chan, "prepaid-final", chan->language);
			if (!res)
				res = ast_waitstream(chan, "");
			res = -1;
		}
	
		return res;
	} else {
		return -1;
	}
}

static int prepaid_ivr_authorize(struct ast_channel *chan, struct calldata *cd)
{
	int res=0;
	int retries;
	// char destination[256] = "";
	char *prompt = NULL;
	char sqlcmd[2048] = "";
	PGresult *result;

  if (check_connected()) {
		// ask destination
		prompt = "prepaid-enter-dest";
		for (retries = 0; retries < 3; retries++) {
			res = ast_app_getdata(chan, prompt, cd->destination, sizeof(cd->destination) - 2, 0);
			if (res < 0)
				break;
			res = 0;
	
			// lookup rate
			sprintf(sqlcmd,psqlauthorization,cd->tariffid,cd->destination);
			ast_log(LOG_DEBUG,"app_prepaid: SQL Authorize command as follows:  %s\n",sqlcmd);
			result = PQexec(conn, sqlcmd);
			if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
				prompt="prepaid-no-aaa"; // tech error sql failed
				//reenter dest
				continue;
			}
			if(PQntuples(result)<=0) {
				prompt="prepaid-dest-unreachable";
				continue;
			}
			cd->rate = PQgetvalue(result,0,0);
			if ( !cd->rate || !strlen(cd->rate) ) {
				prompt="prepaid-dest-blocked";
				continue;
			}
			// TODO: Check values before atoi
			if (!cd->tech) {
				cd->tech = (char *)malloc(strlen(PQgetvalue(result,0,3))+1);
				strcpy (cd->tech,PQgetvalue(result,0,3));
			}
			if (!cd->prefix) {
				cd->prefix = (char *)malloc(strlen(PQgetvalue(result,0,2))+1);
				strcpy (cd->prefix,PQgetvalue(result,0,2));
			}
			if (!cd->hostname) {
				cd->hostname = (char *)malloc(strlen(PQgetvalue(result,0,4))+1);
				strcpy (cd->hostname,PQgetvalue(result,0,4));
			}
			if (PQgetvalue(result,0,5)) {
				cd->providerid = atoi(PQgetvalue(result,0,5));
			} else {
				continue;
			}
			if (PQgetvalue(result,0,6)) {
				cd->countryprefixid = atoi(PQgetvalue(result,0,6));
			} else {
				continue;
			}
			if (PQgetvalue(result,0,1)) {
				cd->tariffrateid = atoi(PQgetvalue(result,0,1));
			} else {
				continue;
			}
			PQclear(result);
	
			ast_log(LOG_DEBUG,"app_prepaid: SQL Authorize results:  rate='%s', tech='%s', prefix='%s', hostname='%s'\n",cd->rate, cd->tech, cd->prefix, cd->hostname);
	
			break;
		}
		if ((retries < 3) && !res) {
			div_t result;
	
			ast_cdr_setapp(chan->cdr, app, cd->destination);
	
			// calculate timeout
			cd->timeout = cd->credit * 60 / atoi(cd->rate);
			ast_log(LOG_DEBUG,"app_prepaid: timeout='%i'",cd->timeout);
	
			// set destination and timeout
			// say 'you have x minutes and x seconds'
	
			result = div (cd->timeout,60);
	
			ast_log(LOG_DEBUG,"app_prepaid: minutes='%i', seconds='%i'",result.quot, result.rem);
	
			res = ast_streamfile(chan, "prepaid-you-have", chan->language);
			if (!res)
				res = ast_waitstream(chan, "");
	
			if (result.quot > 0) {
				res = ast_say_number(chan, result.quot, "", chan->language, "f");
				if (!res)
					res = ast_waitstream(chan, "");
			}
	
			if ( result.quot == 1 ) {
				res = ast_streamfile(chan, "prepaid-minute", chan->language);
				if (!res)
					res = ast_waitstream(chan, "");
	
			} else if (result.quot > 1) {
				res = ast_streamfile(chan, "prepaid-minutes", chan->language);
				if (!res)
					res = ast_waitstream(chan, "");
	
			}
	
			if (result.quot > 0 && result.rem > 0) {
				res = ast_streamfile(chan, "prepaid-and", chan->language);
				if (!res)
					res = ast_waitstream(chan, "");
			}
	
			if (result.rem > 0) {
				if (!res)
					res = ast_waitstream(chan, "");
				res = ast_say_number(chan, result.rem, "", chan->language, "f");
				if (!res)
					res = ast_waitstream(chan, "");
				res = ast_streamfile(chan, "prepaid-seconds", chan->language);
				if (!res)
					res = ast_waitstream(chan, "");
				res = 0;
			}
	
		} else {
			// auth failed
			res = ast_streamfile(chan, "prepaid-final", chan->language);
			if (!res)
				res = ast_waitstream(chan, "");
			res = -1;
		}
	
		return res;
	} else {
		return -1;
	}
}

static int prepaid_pbx_dial(struct ast_channel *chan, void *data)
{
	struct ast_app *app;
	int ret;

	app = pbx_findapp("Dial");
	if (app) {
		ret = pbx_exec(chan, app, data, 1);
	} else {
		ast_log(LOG_WARNING, "Could not find application (Dial)\n");
		ret = -2;
	}
	return ret;
}

static int prepaid_acct_start(struct ast_cdr *cdr, struct calldata *cd)
{
	struct tm tm;
	struct timeval tv;
	char sqlcmd[2048], timestr[128];
	time_t t;
	PGresult *result;

	memset(sqlcmd,0,2048);

	gettimeofday(&tv,NULL);
	t = tv.tv_sec;
	localtime_r(&t,&tm);
	strftime(timestr,128,DATE_FORMAT,&tm);

	if (check_connected()) {
		char *channel=NULL, *lastdata=NULL, *uniqueid=NULL;

		/* Maximum space needed would be if all characters needed to be escaped, plus a trailing NULL */
		if ((lastdata = alloca(strlen(cdr->lastdata) * 2 + 1)) != NULL)
			PQescapeString(lastdata, cdr->lastdata, strlen(cdr->lastdata));
		if ((channel = alloca(strlen(cdr->channel) * 2 + 1)) != NULL)
			PQescapeString(channel, cdr->channel, strlen(cdr->channel));
		if ((uniqueid = alloca(strlen(cdr->uniqueid) * 2 + 1)) != NULL)
			PQescapeString(uniqueid, cdr->uniqueid, strlen(cdr->uniqueid));

		/* Check for all alloca failures above at once */
		if ((!channel) || (!lastdata) || (!uniqueid)) {
			ast_log(LOG_ERROR, "app_prepaid:  Out of memory error (insert fails)\n");
			// ast_mutex_unlock(&prepaid_cdr_lock);
			return -1;
		}

		ast_log(LOG_DEBUG,"app_prepaid: Inserting Prepaid Call record.\n");
		
		sprintf(sqlcmd,psqlacctstart,uniqueid, channel, cd->cardid, timestr, cd->providerid, cd->countryprefixid, cd->tariffrateid, lastdata );

		ast_log(LOG_DEBUG,"app_prepaid: SQL command as follows:  %s\n",sqlcmd);

		result = PQexec(conn, sqlcmd);
		if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
			ast_log(LOG_ERROR,"app_prepaid: Failed to insert accounting record.");
			return -1;
		}
		PQclear(result);
	}
	return 0;
}

static int prepaid_acct_stop(struct ast_cdr *cdr, struct calldata *cd)
{
	struct tm tm;
	struct timeval tv;
	char sqlcmd[2048], timestr[128];
	time_t t;
	PGresult *result;

	memset(sqlcmd,0,2048);

	gettimeofday(&tv,NULL);
	t = tv.tv_sec;
	localtime_r(&t,&tm);
	strftime(timestr,128,DATE_FORMAT,&tm);

	ast_cdr_end(cdr);

	if (!cdr->end.tv_sec && !cdr->end.tv_usec)
		ast_log(LOG_WARNING, "app_prepaid: CDR on channel '%s' lacks end\n", cdr->channel);
	if (!cdr->start.tv_sec && !cdr->start.tv_usec)
		ast_log(LOG_WARNING, "app_prepaid: CDR on channel '%s' lacks start\n", cdr->channel);
	cdr->duration = cdr->end.tv_sec - cdr->start.tv_sec + (cdr->end.tv_usec - cdr->start.tv_usec) / 1000000;
	if (cdr->answer.tv_sec || cdr->answer.tv_usec) {
		cdr->billsec = cdr->end.tv_sec - cdr->answer.tv_sec + (cdr->end.tv_usec - cdr->answer.tv_usec) / 1000000;
	} else
		cdr->billsec = 0;

	if (check_connected()) {
		char *uniqueid=NULL;

		/* Maximum space needed would be if all characters needed to be escaped, plus a trailing NULL */
		if ((uniqueid = alloca(strlen(cdr->uniqueid) * 2 + 1)) != NULL)
			PQescapeString(uniqueid, cdr->uniqueid, strlen(cdr->uniqueid));

		/* Check for all alloca failures above at once */
		if (!uniqueid) {
			ast_log(LOG_ERROR, "app_prepaid:  Out of memory error (insert fails)\n");
			// ast_mutex_unlock(&prepaid_cdr_lock);
			return -1;
		}

		ast_log(LOG_DEBUG,"app_prepaid: Updating Prepaid Call record.\n");

		sprintf(sqlcmd,psqlacctstop,uniqueid, cdr->billsec, ast_cdr_disp2str(cdr->disposition), timestr);

		result = PQexec(conn, sqlcmd);
		if ( PQresultStatus(result) != PGRES_TUPLES_OK) {
			ast_log(LOG_ERROR,"app_prepaid: Failed to update accounting record.");
			return -1;
		}
		PQclear(result);
	}

	return 0;
}

static int prepaid_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	struct localuser *u;
	struct calldata cd;
	char dialstr[256] = "";

	LOCAL_USER_ADD(u);
	if (chan->_state != AST_STATE_UP) {
		res = ast_answer(chan);
		if (res) {
			LOCAL_USER_REMOVE(u);
			return -1;
		}
	}
	
	cd.cardid = NULL;
	cd.tech = NULL;
	cd.prefix = NULL;
	cd.hostname = NULL;
	
	res = prepaid_ivr_authenticate(chan, &cd);
	if (!res) {


		res = prepaid_ivr_authorize(chan, &cd);
		
		prepaid_acct_start(chan->cdr, &cd);
		
		if (!res) {
			ast_cdr_reset(chan->cdr,0);
			ast_channel_setwhentohangup(chan,cd.timeout);

			sprintf(dialstr,"%s/@%s:%s%s", cd.tech, cd.hostname, cd.prefix, chan->cdr->lastdata);
			ast_log(LOG_DEBUG,"app_prepaid: Dialing '%s' with timeout of '%d'.\n",dialstr,cd.timeout);
			res = prepaid_pbx_dial(chan, dialstr);

		}
		prepaid_acct_stop(chan->cdr, &cd);
	}
  if (cd.cardid) 
		free(cd.cardid);
	if (cd.tech)
		free(cd.tech);
	if (cd.prefix)
		free(cd.prefix);
	if (cd.hostname)
		free(cd.hostname);
	LOCAL_USER_REMOVE(u);
	return res;
}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
	PQfinish(conn);
	connected = 0;
	if (pghostname && hostname_alloc) {
		free(pghostname);
		pghostname = NULL;
		hostname_alloc = 0;
	}
	if (pgdbname && dbname_alloc) {
		free(pgdbname);
		pgdbname = NULL;
		dbname_alloc = 0;
	}
	if (pgdbuser && dbuser_alloc) {
		free(pgdbuser);
		pgdbuser = NULL;
		dbuser_alloc = 0;
	}
	if (pgdbsock && dbsock_alloc) {
		free(pgdbsock);
		pgdbsock = NULL;
		dbsock_alloc = 0;
	}
	if (pgpassword && password_alloc) {
		free(pgpassword);
		pgpassword = NULL;
		password_alloc = 0;
	}
	if (pgdbport && dbport_alloc) {
		free(pgdbport);
		pgdbport = NULL;
		dbport_alloc = 0;
	}
	if (psqlauthentication && sqlauthentication_alloc) {
		free(psqlauthentication);
		psqlauthentication = NULL;
		sqlauthentication_alloc = 0;
	}
	if (psqlcheckforsession && sqlcheckforsession_alloc) {
		free(psqlcheckforsession);
		psqlcheckforsession = NULL;
		sqlcheckforsession_alloc = 0;
	}
	if (psqlauthorization && sqlauthorization_alloc) {
		free(psqlauthorization);
		psqlauthorization = NULL;
		sqlauthorization_alloc = 0;
	}
	if (psqlacctstart && sqlacctstart_alloc) {
		free(psqlacctstart);
		psqlacctstart = NULL;
		sqlacctstart_alloc = 0;
	}
	if (psqlacctstop && sqlacctstop_alloc) {
		free(psqlacctstop);
		psqlacctstop = NULL;
		sqlacctstop_alloc = 0;
	}
	if (psqlsimulcount && sqlsimulcount_alloc) {
		free(psqlsimulcount);
		psqlsimulcount = NULL;
		sqlsimulcount_alloc = 0;
	}

	ast_unregister_application(app);

	return 0;
}

int load_module(void)
{
	int res;
	struct ast_config *cfg;
	struct ast_variable *var;
	char *tmp;

	cfg = ast_load(config);
	if (!cfg) {
		ast_log(LOG_WARNING, "Unable to load config for Prepaid Application: %s\n", config);
		return 0;
	}

	var = ast_variable_browse(cfg, "global");
	if (!var) {
		/* nothing configured */
		return 0;
	}

	tmp = ast_variable_retrieve(cfg,"global","hostname");
	if (tmp) {
		pghostname = malloc(strlen(tmp) + 1);
		if (pghostname != NULL) {
			hostname_alloc = 1;
			strcpy(pghostname,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid database server hostname not specified.  Assuming localhost\n");
		pghostname = "localhost";
	}

	tmp = ast_variable_retrieve(cfg,"global","dbname");
	if (tmp) {
		pgdbname = malloc(strlen(tmp) + 1);
		if (pgdbname != NULL) {
			dbname_alloc = 1;
			strcpy(pgdbname,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid database not specified.  Assuming asteriskcdrdb\n");
		pgdbname = "asteriskcdrdb";
	}

	tmp = ast_variable_retrieve(cfg,"global","user");
	if (tmp) {
		pgdbuser = malloc(strlen(tmp) + 1);
		if (pgdbuser != NULL) {
			dbuser_alloc = 1;
			strcpy(pgdbuser,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid database user not specified.  Assuming root\n");
		pgdbuser = "root";
	}

	tmp = ast_variable_retrieve(cfg,"global","password");
	if (tmp) {
		pgpassword = malloc(strlen(tmp) + 1);
		if (pgpassword != NULL) {
			password_alloc = 1;
			strcpy(pgpassword,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid database password not specified.  Assuming blank\n");
		pgpassword = "";
	}

	tmp = ast_variable_retrieve(cfg,"global","port");
	if (tmp) {
		pgdbport = malloc(strlen(tmp) + 1);
		if (pgdbport != NULL) {
			dbport_alloc = 1;
			strcpy(pgdbport,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid database port not specified.  Using default.\n");
		pgdbport = "5432";
	}


	tmp = ast_variable_retrieve(cfg,"sql","authentication");
	if (tmp) {
		psqlauthentication = malloc(strlen(tmp) + 1);
		if (psqlauthentication != NULL) {
			sqlauthentication_alloc = 1;
			strcpy(psqlauthentication,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid authentication query not specified.  Using default.\n");
		psqlauthentication = "SELECT * FROM asterisk_authenticate('%s') AS authenticate(credit integer, tariffid integer, activated bool, expired bool, cardid text)";
	}

	tmp = ast_variable_retrieve(cfg,"sql","checkforsession");
	if (tmp) {
		psqlcheckforsession = malloc(strlen(tmp) + 1);
		if (psqlcheckforsession != NULL) {
			sqlcheckforsession_alloc = 1;
			strcpy(psqlcheckforsession,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid checkforsession query not specified.  Using default.\n");
		psqlcheckforsession = "SELECT stoptime is NULL FROM card left join cid USING(cidid) left join call USING(cardid) WHERE stoptime is NULL and cid='%s' and not (starttime is NULL)";
	}
	
	tmp = ast_variable_retrieve(cfg,"sql","authorization");
	if (tmp) {
		psqlauthorization = malloc(strlen(tmp) + 1);
		if (psqlauthorization != NULL) {
			sqlauthorization_alloc = 1;
			strcpy(psqlauthorization,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid authorization query not specified.  Using default.\n");
		psqlauthorization = "SELECT * FROM asterisk_authorize(%d,'%s') AS authorize(rate integer, tariffrateid integer, prefix text, tech text, ipaddress text, providerid integer, countryprefixid integer)";
	}


	tmp = ast_variable_retrieve(cfg,"sql","acct_start");
	if (tmp) {
		psqlacctstart = malloc(strlen(tmp) + 1);
		if (psqlacctstart != NULL) {
			sqlacctstart_alloc = 1;
			strcpy(psqlacctstart,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid acct_start query not specified.  Using default.\n");
		psqlacctstart = "SELECT asterisk_acct_start('%s','%s','%s','%s',%d,%d,%d,'%s')";
	}

	tmp = ast_variable_retrieve(cfg,"sql","acct_stop");
	if (tmp) {
		psqlacctstop = malloc(strlen(tmp) + 1);
		if (psqlacctstop != NULL) {
			sqlacctstop_alloc = 1;
			strcpy(psqlacctstop,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid acct_stop query not specified.  Using default.\n");
		psqlacctstop = "SELECT asterisk_acct_stop('%s','%i','%s','%s')";
	}

	tmp = ast_variable_retrieve(cfg,"sql","simulcount");
	if (tmp) {
		psqlsimulcount = malloc(strlen(tmp) + 1);
		if (psqlsimulcount != NULL) {
			sqlsimulcount_alloc = 1;
			strcpy(psqlsimulcount,tmp);
		} else {
			ast_log(LOG_ERROR,"Out of memory error.\n");
			return -1;
		}
	} else {
		ast_log(LOG_WARNING,"Prepaid simulcount query not specified.  Using default.\n");
		psqlsimulcount = "select count(*) from call where username='%s' and stoptime is null;";
	}

	ast_destroy(cfg);

	ast_log(LOG_DEBUG,"app_prepaid: got hostname of %s\n",pghostname);
	ast_log(LOG_DEBUG,"app_prepaid: got port of %s\n",pgdbport);
	if (pgdbsock)
		ast_log(LOG_DEBUG,"app_prepaid: got sock file of %s\n",pgdbsock);
	ast_log(LOG_DEBUG,"app_prepaid: got user of %s\n",pgdbuser);
	ast_log(LOG_DEBUG,"app_prepaid: got dbname of %s\n",pgdbname);
	ast_log(LOG_DEBUG,"app_prepaid: got password of %s\n",pgpassword);

	ast_log(LOG_DEBUG,"app_prepaid: got sqlauthentication of '%s'\n",psqlauthentication);
	ast_log(LOG_DEBUG,"app_prepaid: got sqlcheckforsession of '%s'\n",psqlcheckforsession);
	ast_log(LOG_DEBUG,"app_prepaid: got sqlauthorization of '%s'\n",psqlauthorization);
	ast_log(LOG_DEBUG,"app_prepaid: got sqlacct_start of '%s'\n",psqlacctstart);
	ast_log(LOG_DEBUG,"app_prepaid: got sqlacct_stop of '%s'\n",psqlacctstop);
	ast_log(LOG_DEBUG,"app_prepaid: got sqlsimulcount of '%s'\n",psqlsimulcount);

	// We are not connected
	connected = 0;
	// so, check connection (it will automatically set up the connection)
	check_connected();

	res = ast_register_application(app, prepaid_exec, synopsis, descrip);
	if (res) {
		ast_log(LOG_ERROR, "Unable to register PrepaidCID Application\n");
	}

	return res;
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}

